#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include <stdbool.h>
#include <ncurses.h>
#include <unistd.h>
#include <termios.h>
#include <limits.h>
#include <time.h>
#include <glad/glad.h>
#include <GLFW/glfw3.h>

#include "hash.h"
#include "structure.h"
#include "base.h"
#include "move.h"
#include "proj.h"
#include "entities.h"
#include "display.h"
#include "generation.h"

double sim_time ;

void processInput(GLFWwindow *window, float dtime) {
    if (glfwGetKey(window, GLFW_KEY_ESCAPE) == GLFW_PRESS) {
        glfwSetWindowShouldClose(window, true);
    }
    if(glfwGetKey(window, GLFW_KEY_S) == GLFW_PRESS) {
        for(int k = 0; k < 10; k++) {
            camz -= speed*cos(rot_hz)*cos(rot_vt)/10;
            camx -= speed*sin(rot_hz)*cos(rot_vt)/10;
            camy += speed*sin(rot_vt)/10;
            if(is_colliding(dtime)) {
                camz += speed*cos(rot_hz)*cos(rot_vt)/10;
                camx += speed*sin(rot_hz)*cos(rot_vt)/10;
                camy -= speed*sin(rot_vt)/10;
                k=11;
            }
        }
    }
    if(glfwGetKey(window, GLFW_KEY_D) == GLFW_PRESS) {
        for(int k = 0; k < 10; k++) {
            camx -= speed*cos(rot_hz)/10;
            camz += speed*sin(rot_hz)/10;
            if(is_colliding(dtime)) {
                camx += speed*cos(rot_hz)/10;
                camz -= speed*sin(rot_hz)/10;
                k=11;
            }
        }
    }
    if(glfwGetKey(window, GLFW_KEY_Z) == GLFW_PRESS) {
        for(int k = 0; k < 10; k++) {
            camz += speed*cos(rot_hz)*cos(rot_vt)/10;
            camx += speed*sin(rot_hz)*cos(rot_vt)/10;
            camy -= speed*sin(rot_vt)/10;
            if(is_colliding(dtime)) {
                camz -= speed*cos(rot_hz)*cos(rot_vt)/10;
                camx -= speed*sin(rot_hz)*cos(rot_vt)/10;
                camy += speed*sin(rot_vt)/10;
                k=11;
            }
        }
    }
    if(glfwGetKey(window, GLFW_KEY_Q) == GLFW_PRESS) {
        for(int k = 0; k < 10; k++) {
            camx += speed*cos(rot_hz)/10;
            camz -= speed*sin(rot_hz)/10;
            if(is_colliding(dtime)) {
                camx -= speed*cos(rot_hz)/10;
                camz += speed*sin(rot_hz)/10;
                k=11;
            }
        }
    }

    if(glfwGetKey(window, GLFW_KEY_SPACE) == GLFW_PRESS) {
        stop_evetything = true ;
    }

    if(glfwGetKey(window, GLFW_KEY_E) == GLFW_PRESS) {
        rot_hz -= sensitivity;
    }
    if(glfwGetKey(window, GLFW_KEY_A) == GLFW_PRESS) {
        rot_hz += sensitivity;
    }

    if(glfwGetKey(window, GLFW_KEY_P) == GLFW_PRESS) {
        rot_vt -= sensitivity;
    }
    if(glfwGetKey(window, GLFW_KEY_M) == GLFW_PRESS) {
        rot_vt += sensitivity;
    }
}

const char *vertexShaderSource = "#version 330 core\n"
    "layout (location = 0) in vec3 aPos;\n"
    "uniform mat4 scale;\n"
    "uniform mat4 model;\n"
    "uniform mat4 view;\n"
    "uniform mat4 projection;\n"
    "void main() {\n"
    "   gl_Position = projection * view * model * scale * vec4(aPos, 1.0);\n"
    "}\0";

// Fragment Shader Source
const char *fragmentShaderSource = "#version 330 core\n"
    "uniform vec4 u_color;\n"
    "out vec4 FragColor;\n"
    "void main() {\n"
    "   FragColor = u_color;\n"
    "}\0";

int main_alt() {
    // glfw: initialize and configure
    // ------------------------------
    glfwInit();
    glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
    glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
    glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);

    // glfw window creation
    // --------------------
    GLFWwindow* window = glfwCreateWindow(1500, 1000, "LearnOpenGL", NULL, NULL);
    if (window == NULL)
    {
        fprintf(stderr, "ERROR : cannot initialize GL window");
        glfwTerminate();
        return -1;
    }
    glfwMakeContextCurrent(window);

    // glad: load all OpenGL function pointers
    // ---------------------------------------
    if (!gladLoadGLLoader((GLADloadproc)glfwGetProcAddress))
    {
        fprintf(stderr, "ERROR : cannot initialize GLAD loader");
        return -1;
    }

    glEnable(GL_DEPTH_TEST);
    glEnable(GL_CULL_FACE);
    glCullFace(GL_FRONT); 
    glDepthFunc(GL_LESS);

    //printf("%f\n", glDepthRange);

    // build and compile our shader program
    // ------------------------------------
    // vertex shader
    unsigned int vertexShader = glCreateShader(GL_VERTEX_SHADER);
    glShaderSource(vertexShader, 1, &vertexShaderSource, NULL);
    glCompileShader(vertexShader);
    // check for shader compile errors
    int success;
    char infoLog[512];
    glGetShaderiv(vertexShader, GL_COMPILE_STATUS, &success);
    if (!success)
    {
        glGetShaderInfoLog(vertexShader, 512, NULL, infoLog);
        fprintf(stderr, "ERROR : cannot initialize shader (1)");
    }
    // fragment shader
    unsigned int fragmentShader = glCreateShader(GL_FRAGMENT_SHADER);
    glShaderSource(fragmentShader, 1, &fragmentShaderSource, NULL);
    glCompileShader(fragmentShader);
    // check for shader compile errors
    glGetShaderiv(fragmentShader, GL_COMPILE_STATUS, &success);
    if (!success)
    {
        glGetShaderInfoLog(fragmentShader, 512, NULL, infoLog);
        fprintf(stderr, "ERROR : cannot initialize shader (2)");
    }
    // link shaders
    unsigned int shaderProgram = glCreateProgram();
    glAttachShader(shaderProgram, vertexShader);
    glAttachShader(shaderProgram, fragmentShader);
    glLinkProgram(shaderProgram);
    // check for linking errors
    glGetProgramiv(shaderProgram, GL_LINK_STATUS, &success);
    if (!success) {
        glGetProgramInfoLog(shaderProgram, 512, NULL, infoLog);
        fprintf(stderr, "ERROR : cannot link shaders");
    }

    glDeleteShader(vertexShader);
    glDeleteShader(fragmentShader);

    // set up data (buffer(s)) and configure vertex attributes
    // ------------------------------------------------------------------

    init_csts();
    init_hashtbl();
    init_ent_generator(10);
    init_proj();
    parse_rooms(5);

    int fps = 60 ;
    int interval = 1000000/fps ;

    clock_t finish = clock();
    clock_t origin = clock();

    unsigned int VBO, VAO;
    glGenVertexArrays(1, &VAO);
    glGenBuffers(1, &VBO);

    glBindVertexArray(VAO);

    init_vertices();

    glBindBuffer(GL_ARRAY_BUFFER, VBO);
    glBufferData(GL_ARRAY_BUFFER, sizeof(vertices), vertices, GL_DYNAMIC_DRAW);

    // position attribute
    glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 3*sizeof(float), (void*)0);
    glEnableVertexAttribArray(0);

    while(!glfwWindowShouldClose(window)) {
        // input
        // -----
        glClearColor(0.2f, 0.3f, 0.3f, 1.0f);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        origin = clock();

        generate_nearby_chunks(1);

        glBindVertexArray(VAO);
        glBindBuffer(GL_ARRAY_BUFFER, VBO);
        gl_initRender(shaderProgram, shaderProgram, VAO, VBO);
        //gl_renderAll(shaderProgram, shaderProgram, VAO, VBO, current_room, 0.0, 0.0, 0.0);
        gl_renderNearbyChunks(shaderProgram, shaderProgram, VAO, VBO, 1);
        gl_renderProj(shaderProgram, shaderProgram, VAO, VBO);

        // glfw: swap buffers and poll IO events (keys pressed/released, mouse moved etc.)
        // -------------------------------------------------------------------------------
        finish = clock();
        processInput(window, ((float)finish - (float)origin)/CLOCKS_PER_SEC);
        teleport_on_edge();
        update_entities(((float)finish - (float)origin)/CLOCKS_PER_SEC);
        updateProj(((float)finish - (float)origin)/CLOCKS_PER_SEC);
        printf("%f\n", 1.0f/(((float)finish - (float)origin)/CLOCKS_PER_SEC));
        //printf("%lf, %lf, %lf\n", camx, camy, camz);

        usleep(interval);

        glfwSwapBuffers(window);
        glfwPollEvents();
    }

    //printf("10\n");
    //fflush(stdout);
    hashtbl_free(visited);
    free_proj();
    free_pool();

    // optional: de-allocate all resources once they've outlived their purpose:
    // ------------------------------------------------------------------------
    glDeleteVertexArrays(1, &VAO);
    glDeleteBuffers(1, &VBO);
    glDeleteProgram(shaderProgram);

    // glfw: terminate, clearing all previously allocated GLFW resources.
    // ------------------------------------------------------------------
    glfwTerminate();
    return 0;
}

int main(int argc, char** argv) {
    srand(time(NULL));
    return main_alt();
}